#!/usr/bin/env python
# -*- coding: iso-8859-15 -*-
# generated by wxGlade 0.6.3 on Thu Jul 17 13:06:16 2008
#
# RemoteKB
#
# Copyright (c) 2008 Arkadiusz Wahlig
# <arkadiusz.wahlig@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import wx, wx.lib.dialogs
from ConfigParser import SafeConfigParser
import os, threading, time
from key_codes import *

__version__ = '1.00.2'

# begin wxGlade: extracode
# end wxGlade

class Keyboard(object):
    # key simulation types, see remote() method
    TYPE_SIMULATE_KEY = 0 # args: keycode, scancode, [modifiers]
    TYPE_SIMULATE_EVENT_UP = 1 # args: keycode, scancode, [modifiers]
    TYPE_SIMULATE_EVENT_DOWN = 2 # args: keycode, scancode, [modifiers]
    TYPE_SIMULATE_EVENT_KEY = 3 # args: keycode, scancode, [modifiers]
    TYPE_SEND_RAW_EVENT_UP = 4 # args: scancode
    TYPE_SEND_RAW_EVENT_DOWN = 5 # args: scancode
    
    # used with TYPE_SEND_RAW_EVENT_xxx
    map_to_scancode = {
        # local keycode: remote scancode
        wx.WXK_SHIFT: EStdKeyLeftShift,
        wx.WXK_ALT: EStdKeyLeftAlt,
        wx.WXK_CONTROL: EStdKeyLeftCtrl,
        wx.WXK_CAPITAL: EStdKeyCapsLock,
        wx.WXK_F1: EScancodeLeftSoftkey,
        wx.WXK_F1|0x8000: EStdKeyApplication9, # (Ctrl+F1 -> Portrait)
        wx.WXK_F2: EScancodeSelect,
        wx.WXK_F2|0x8000: EStdKeyApplicationA, # (Ctrl+F2 -> Landscape)
        wx.WXK_F3: EScancodeRightSoftkey,
        wx.WXK_F4: EStdKeyApplication0, # (Application/Menu)
        wx.WXK_F5: EStdKeyYes,
        wx.WXK_F6: EStdKeyNo,
        wx.WXK_F6|0x8000: EStdKeyDevice2, # (Ctrl+F6 -> Power)
        wx.WXK_F7: EStdKeyDecVolume,
        wx.WXK_F8: EStdKeyIncVolume,
        wx.WXK_F9: EStdKeyApplication2, # (Play/Pause)
        wx.WXK_F10: EStdKeyApplication3, # (Stop)
        wx.WXK_F11: EStdKeyApplication5, # (Rewind)
        wx.WXK_F11|0x8000: EScancodeStar, # (Ctrl+F11 -> Star)
        wx.WXK_F12: EStdKeyApplication4, # (Forward)
        wx.WXK_F12|0x8000: EStdKeyHash, # (Ctrl+F12 -> Hash)
        wx.WXK_PAGEUP: EStdKeyPageUp,
        wx.WXK_PAGEDOWN: EStdKeyPageDown,
        wx.WXK_NUMPAD0: EScancode0,
        wx.WXK_NUMPAD1: EScancode1,
        wx.WXK_NUMPAD2: EScancode2,
        wx.WXK_NUMPAD3: EScancode3,
        wx.WXK_NUMPAD4: EScancode4,
        wx.WXK_NUMPAD5: EScancode5,
        wx.WXK_NUMPAD6: EScancode6,
        wx.WXK_NUMPAD7: EScancode7,
        wx.WXK_NUMPAD8: EScancode8,
        wx.WXK_NUMPAD9: EScancode9,
        wx.WXK_NUMPAD_MULTIPLY: EScancodeStar,
        wx.WXK_NUMPAD_DECIMAL: EStdKeyHash,
    }
    
    # used with TYPE_SIMULATE_EVENT_xxx
    map_to_key_and_scancode = {
        wx.WXK_BACK: (EKeyBackspace, EStdKeyBackspace),
        wx.WXK_TAB: (EKeyTab, EStdKeyTab),
        wx.WXK_RETURN: (EKeyEnter, EStdKeyEnter),
        wx.WXK_ESCAPE: (EKeyEscape, EStdKeyEscape),
        wx.WXK_SPACE: (EKeySpace, EStdKeySpace),
        wx.WXK_DELETE: (EKeyDelete, EStdKeyDelete),
        wx.WXK_PAUSE: (EKeyPause, EStdKeyPause),
        wx.WXK_END: (EKeyEnd, EStdKeyEnd),
        wx.WXK_HOME: (EKeyHome, EStdKeyHome),
        wx.WXK_INSERT: (EKeyInsert, EStdKeyInsert),
        wx.WXK_LEFT: (EKeyLeftArrow, EStdKeyLeftArrow),
        wx.WXK_UP: (EKeyUpArrow, EStdKeyUpArrow),
        wx.WXK_RIGHT: (EKeyRightArrow, EStdKeyRightArrow),
        wx.WXK_DOWN: (EKeyDownArrow, EStdKeyDownArrow),
    }
    
    # maps local modifiers to remote ones
    map_modifier = {
        # local modifier: remote modifiers
        wx.MOD_ALT: EModifierLeftAlt|EModifierAlt,
        wx.MOD_CONTROL: EModifierLeftCtrl|EModifierCtrl,
        wx.MOD_SHIFT: EModifierLeftShift|EModifierShift,
    }
    
    eat_keycodes = [
        wx.WXK_F10, wx.WXK_ALT, wx.WXK_NUMPAD0, wx.WXK_NUMPAD1, wx.WXK_NUMPAD2,
        wx.WXK_NUMPAD3, wx.WXK_NUMPAD4, wx.WXK_NUMPAD5, wx.WXK_NUMPAD6,
        wx.WXK_NUMPAD7, wx.WXK_NUMPAD8, wx.WXK_NUMPAD9, wx.WXK_NUMPAD_DECIMAL,
        wx.WXK_NUMPAD_MULTIPLY,
    ]
    
    def __init__(self):
        self.porth = None
        self.port = ''
        self.pressed = []
        self.last_key = (), 0.0
        self.arrows_shift_mode = False
        self.shift_down = False
    
    def open(self, port):
        if self.porth:
            if self.port == port:
                return
            self.close()
        self.porth = open(port, 'w')
        self.port = port
            
    def close(self):
        try:
            self.porth.close()
        except:
            pass
        self.porth = None
        self.port = ''

    def set_arrows_shift_mode(self, on):
        self.arrows_shift_mode = on

    def on_key_event(self, event):
        '''Bind this to wx.EVT_KEY_DOWN, wx.EVT_KEY_UP and wx.EVT_CHAR.
        '''
        if not self.porth:
            raise IOError('port is not opened')
            
        type = event.GetEventType()
        keycode = event.GetKeyCode()
        modifiers = event.GetModifiers()

        if keycode not in self.eat_keycodes:
            event.Skip()

        if keycode == wx.WXK_SHIFT and event.GetRawKeyCode() == 0x10 and \
                event.GetRawKeyFlags()&0xfffffff == 0x100001:
            # some strange shift keypresses out of nowhere
            return

        #print 'type=%s, keycode=%x, modifiers=%x, raw=%x, rawflags=%x, unichar=%s, unikey=%s' % (event.GetEventType(), event.GetKeyCode(), event.GetModifiers(), event.GetRawKeyCode(), event.GetRawKeyFlags(), repr(unichr(event.GetUniChar())), repr(unichr(event.GetUnicodeKey())))

        self.process_key(type, keycode, event.GetModifiers(), event.GetUnicodeKey())
        
    def process_key(self, type, keycode, modifiers, unicodekey=0):
        code = keycode
        if modifiers & 2: # ctrl
            code |= 0x8000
        # remote modifiers
        rmodifiers = 0
        for m, rm in self.map_modifier.items():
            if modifiers & m:
                rmodifiers |= rm
        # used to slow down keyrepeat
        curtime = time.time()

        if type == wx.EVT_KEY_DOWN.typeId:
            try:
                scancode = self.map_to_scancode[code]
            except KeyError:
                try:
                    key, scancode = self.map_to_key_and_scancode[code]
                except KeyError:
                    pass
                else:
                    if self.arrows_shift_mode and keycode in \
                            (wx.WXK_UP, wx.WXK_LEFT, wx.WXK_RIGHT, wx.WXK_DOWN) and \
                            wx.WXK_SHIFT in self.pressed and not self.shift_down:
                        self.remote(self.TYPE_SEND_RAW_EVENT_DOWN, self.map_to_scancode[wx.WXK_SHIFT])
                        self.shift_down = True
                    if keycode not in self.pressed:
                        self.remote(self.TYPE_SIMULATE_EVENT_DOWN, 0, scancode, rmodifiers)
                    if self.last_key[0] != (keycode, rmodifiers) or \
                            self.last_key[1] + 0.10 < curtime:
                        self.remote(self.TYPE_SIMULATE_EVENT_KEY, key, scancode, rmodifiers)
                        self.last_key = (keycode, rmodifiers), curtime
            else:
                if keycode not in self.pressed:
                    if keycode == wx.WXK_SHIFT and self.arrows_shift_mode:
                        self.shift_down = False
                    else:
                        self.remote(self.TYPE_SEND_RAW_EVENT_DOWN, scancode)
        elif type == wx.EVT_KEY_UP.typeId:
            try:
                scancode = self.map_to_scancode[code]
            except KeyError:
                try:
                    key, scancode = self.map_to_key_and_scancode[code]
                except KeyError:
                    pass
                else:
                    self.remote(self.TYPE_SIMULATE_EVENT_UP, 0, scancode, rmodifiers)
            else:
                if not self.arrows_shift_mode or keycode != wx.WXK_SHIFT or self.shift_down:
                    self.remote(self.TYPE_SEND_RAW_EVENT_UP, scancode)
        elif type == wx.EVT_CHAR.typeId:
            if code not in self.map_to_scancode and \
                    code not in self.map_to_key_and_scancode:
                if self.last_key[0] != (keycode, rmodifiers) or \
                        self.last_key[1] + 0.08 < curtime:
                    # add EModifierKeyboardExtend modifier
                    self.remote(self.TYPE_SIMULATE_KEY, unicodekey, 0, rmodifiers|0x200000)
                    self.last_key = (keycode, rmodifiers), curtime

        if type == wx.EVT_KEY_DOWN.typeId:
            if keycode not in self.pressed:
                self.pressed.append(keycode)
        elif type == wx.EVT_KEY_UP.typeId:
            try:
                self.pressed.remove(keycode)
            except ValueError:
                pass

    def remote(self, type, *args):
        '''Sends a keypress to the phone.
        '''
        print type, args
        self.porth.write('%s\r' % repr((type,)+args))
        self.porth.flush()


class MainFrame(wx.Frame):
    def __init__(self, *args, **kwds):
        # begin wxGlade: MainFrame.__init__
        kwds["style"] = wx.DEFAULT_FRAME_STYLE
        wx.Frame.__init__(self, *args, **kwds)
        self.panel = wx.Panel(self, -1)
        
        # Menu Bar
        self.frame_menubar = wx.MenuBar()
        wxglade_tmp_menu = wx.Menu()
        wxglade_tmp_menu.Append(1, "Active", "", wx.ITEM_CHECK)
        wxglade_tmp_menu_sub = wx.Menu()
        wxglade_tmp_menu_sub.Append(4, "Normal\tAlt+N", "", wx.ITEM_RADIO)
        wxglade_tmp_menu_sub.Append(5, "Arrows only\tAlt+A", "", wx.ITEM_RADIO)
        wxglade_tmp_menu.AppendMenu(wx.NewId(), "Shift mode", wxglade_tmp_menu_sub, "")
        self.frame_menubar.Append(wxglade_tmp_menu, "Keyboard")
        wxglade_tmp_menu = wx.Menu()
        wxglade_tmp_menu.Append(2, "Bluetooth serial port device...", "", wx.ITEM_NORMAL)
        self.frame_menubar.Append(wxglade_tmp_menu, "Tools")
        wxglade_tmp_menu = wx.Menu()
        wxglade_tmp_menu.Append(3, "About...", "", wx.ITEM_NORMAL)
        self.frame_menubar.Append(wxglade_tmp_menu, "Help")
        self.SetMenuBar(self.frame_menubar)
        # Menu Bar end
        self.statictext_copy_copy = wx.StaticText(self.panel, -1, "Activate this window to send\nall keypresses to the phone\nover Bluetooth.", style=wx.ALIGN_CENTRE)
        self.statictext = wx.StaticText(self.panel, -1, "F1 - Left softkey\nF2 - Select (Joystick)\nF3 - Right softkey\nF4 - Application (Menu)\n\nF5 - Green key\nF6 - Red key\nF7 - Volume down\nF8 - Volume up\n\nF9 - Play / Pause\nF10 - Stop\nF11 - Rewind\nF12 - Forward")
        self.statictext_copy = wx.StaticText(self.panel, -1, "Ctrl+F1 - Portrait mode\nCtrl+F2 - Landscape mode\n\nCtrl+F6 - Power\n\nCtrl+F11 - Device key *\nCtrl+F12 - Device key #\n\nNumpad 0-9 - Device keys 0-9\nNumpad * - Device key *\nNumpad . - Device key #")

        self.__set_properties()
        self.__do_layout()

        self.Bind(wx.EVT_MENU, self.on_active, id=1)
        self.Bind(wx.EVT_MENU, self.on_shift_mode_normal, id=4)
        self.Bind(wx.EVT_MENU, self.on_shift_mode_arrows_only, id=5)
        self.Bind(wx.EVT_MENU, self.on_serial_port, id=2)
        self.Bind(wx.EVT_MENU, self.on_about, id=3)
        # end wxGlade

        from sys import platform
        if platform == 'win32':
            defport = 'COM7:'
        elif platform.startswith('linux'):
            defport = '/dev/rfcomm2'
        else:
            defport = ''

        self.parser = SafeConfigParser()
        self.parser.add_section('main')
        self.parser.set('main', 'port', defport)
        self.parser.set('main', 'arrows_shift_mode', 'false')
        path, name = os.path.split(__file__)
        name = os.path.splitext(name)[0]
        self.inipath = os.path.join(path, '%s.ini' % name)
        self.parser.read(self.inipath)
        
        self.keyboard = Keyboard()
        if platform.startswith('linux'):
            self.panel.Bind(wx.EVT_KEY_DOWN, self.on_key, id=-1)
            self.panel.Bind(wx.EVT_CHAR, self.on_key, id=-1)
            self.panel.Bind(wx.EVT_KEY_UP, self.on_key, id=-1)
            self.panel.SetFocus()
        else:
            self.Bind(wx.EVT_KEY_DOWN, self.on_key, id=-1)
            self.Bind(wx.EVT_CHAR, self.on_key, id=-1)
            self.Bind(wx.EVT_KEY_UP, self.on_key, id=-1)

        mode = self.parser.getboolean('main', 'arrows_shift_mode')
        if mode:
            self.frame_menubar.FindItemById(5).Check(True)
            self.keyboard.set_arrows_shift_mode(True)

    def __set_properties(self):
        # begin wxGlade: MainFrame.__set_properties
        self.SetTitle("RemoteKB")
        # end wxGlade

    def __do_layout(self):
        # begin wxGlade: MainFrame.__do_layout
        sizer1 = wx.BoxSizer(wx.VERTICAL)
        sizer2 = wx.BoxSizer(wx.VERTICAL)
        sizer_1 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_2 = wx.BoxSizer(wx.VERTICAL)
        sizer_3 = wx.BoxSizer(wx.HORIZONTAL)
        sizer2.Add((8, 8), 1, wx.EXPAND, 0)
        sizer_1.Add((8, 8), 1, wx.EXPAND, 0)
        sizer_2.Add(self.statictext_copy_copy, 0, wx.ALL|wx.EXPAND, 8)
        sizer_3.Add(self.statictext, 0, wx.ALL, 8)
        sizer_3.Add(self.statictext_copy, 0, wx.ALL|wx.ALIGN_RIGHT, 8)
        sizer_2.Add(sizer_3, 0, 0, 0)
        sizer_1.Add(sizer_2, 0, wx.ALIGN_CENTER_HORIZONTAL|wx.ALIGN_CENTER_VERTICAL, 0)
        sizer_1.Add((8, 8), 1, wx.EXPAND, 0)
        sizer2.Add(sizer_1, 1, wx.EXPAND, 0)
        sizer2.Add((8, 8), 1, wx.EXPAND, 0)
        self.panel.SetSizer(sizer2)
        sizer1.Add(self.panel, 1, wx.EXPAND, 0)
        self.SetSizer(sizer1)
        sizer1.Fit(self)
        self.Layout()
        # end wxGlade

    def on_active(self, event): # wxGlade: MainFrame.<event_handler>
        self.setup_port()

    def on_serial_port(self, event): # wxGlade: MainFrame.<event_handler>
        r = wx.lib.dialogs.textEntryDialog(self, 'Bluetooth serial port device:',
            self.GetTitle(), self.parser.get('main', 'port'))
        if r.returned == wx.ID_OK:
            self.parser.set('main', 'port', str(r.text))
            self.parser.write(open(self.inipath, 'w'))
            self.setup_port()

    def on_about(self, event): # wxGlade: MainFrame.<event_handler>
        wx.MessageBox('RemoteKB-desktop %s\n'
            '\n'
            'Copyright (c) 2008 Arkadiusz Wahlig\n'
            '<arkadiusz.wahlig@gmail.com>' % __version__,
            self.GetTitle())

    def setup_port(self):
        item = self.frame_menubar.FindItemById(1)
        if item.IsChecked():
            port = self.parser.get('main', 'port')
            if port:
                try:
                    self.keyboard.open(port)
                except IOError, e:
                    wx.MessageBox('Cannot open port (%s)\n%s' % (port, e),
                        self.GetTitle())
                    item.Toggle()
            else:
                wx.MessageBox('Port not specified, select\n'
                    'Bluetooth serial port device\n'
                    'from Tools menu.', self.GetTitle())
                item.Toggle()
        else:
            self.keyboard.close()
            
    def on_key(self, event):
        try:
            self.keyboard.on_key_event(event)
        except IOError:
            item = self.frame_menubar.FindItemById(1)
            if not item.IsChecked():
                if event.GetEventType() == wx.EVT_KEY_DOWN.typeId:
                    r = wx.MessageBox('Keyboard is currently deactivated.\n'
                        'Do you want to activate it?', self.GetTitle(), wx.YES_NO)
                    if r is wx.YES:
                        item.Toggle()
                        self.on_active(None)

    def on_shift_mode_normal(self, event): # wxGlade: MainFrame.<event_handler>
        self.keyboard.set_arrows_shift_mode(False)
        self.parser.set('main', 'arrows_shift_mode', 'false')
        self.parser.write(open(self.inipath, 'w'))
        event.Skip()

    def on_shift_mode_arrows_only(self, event): # wxGlade: MainFrame.<event_handler>
        self.keyboard.set_arrows_shift_mode(True)
        self.parser.set('main', 'arrows_shift_mode', 'true')
        self.parser.write(open(self.inipath, 'w'))
        event.Skip()

# end of class MainFrame


if __name__ == "__main__":
    app = wx.PySimpleApp(0)
    wx.InitAllImageHandlers()
    frame = MainFrame(None, -1, "")
    app.SetTopWindow(frame)
    frame.SetSize((400, 350))
    frame.Show()
    frame.frame_menubar.FindItemById(1).Toggle()
    frame.setup_port()
    app.MainLoop()
